# a simple interface to turtle field for joho-kiso
# Yoshinori Hayakawa
# 02-JUL-2013
# 19-JUN-2017 (bug fix)
# 30-JUN-2017

import sys
import socket
import time
import re

BUFFSIZE=8192
PORT=2222
EVENT_HIT_BY_BULLET   = (0x00000001)
EVENT_RUN_INTO_TURTLE = (0x00000002)
EVENT_RUN_INTO_DONUT  = (0x00000004)
EVENT_RUN_INTO_STONE  = (0x00000008)
EVENT_RUN_INTO_WALL   = (0x00000010)
EVENT_FOUND_COIN      = (0x00000020)
EVENT_DETECTED_BY_FINDER = (0x00000040)
EVENT_DETECTED_BY_RADAR  = (0x00000080)
EVENT_GOT_MESSAGE        = (0x00000100)

class Turtle:
    def __init__(self,hostname="localhost"):
        self.socket = socket.socket()
        self.last_message=""
        self.team=0 
        try:
            self.socket.connect((hostname, PORT))
        except IOError:
            print >> sys.stderr,"cannot connect to %s" % hostname
            print >> sys.stderr,"seems like turtle field is not ready..."
            
    def ttl_send_command(self,cmd):
        rc=0
        try:
            self.socket.send(cmd + "\n")
            self.last_message = self.socket.recv(BUFFSIZE)
        except IOError:
            print >> sys.stderr,"communiaction error... bye"
            sys.exit()
        if not self.last_message:
            print >> sys.stderr,"I was killed.. bye"
            sys.exit()
        else:
            time.sleep(0.001)
        ack = self.last_message.split(" ")
        eventcode = int(ack[-1])
        timecode = int(ack[-2])
        if ack[-4]=="OK":
            rc=0
        else:
            rc=-1
        if eventcode != 0 :
            self.check_event(eventcode,timecode)
        return rc

    def enable_event(self,eventcode):
        self.ttl_send_command("Y + %d"%eventcode)

    def disable_event(self,eventcode):
        self.ttl_send_command("Y - %d"%eventcode)

    def check_event(self,event,timecode):
        curr_team = self.team
        msg_stack = self.last_message[:]
        if event & EVENT_HIT_BY_BULLET !=0 :
            self.disable_event(EVENT_HIT_BY_BULLET)
            self.hit_by_bullet(timecode)
            self.enable_event(EVENT_HIT_BY_BULLET)
        if event & EVENT_RUN_INTO_TURTLE != 0 :
            self.disable_event(EVENT_RUN_INTO_TURTLE)
            self.run_into_turtle(timecode)
            self.enable_event(EVENT_RUN_INTO_TURTLE)
        if event & EVENT_RUN_INTO_DONUT != 0 :
            self.disable_event(EVENT_RUN_INTO_DONUT)
            self.run_into_donut(timecode)
            self.enable_event(EVENT_RUN_INTO_DONUT)
        if event & EVENT_RUN_INTO_STONE != 0 :
            self.disable_event(EVENT_RUN_INTO_STONE)
            self.run_into_stone(timecode)
            self.enable_event(EVENT_RUN_INTO_STONE)
        if event & EVENT_RUN_INTO_WALL != 0 :
            self.disable_event(EVENT_RUN_INTO_WALL)
            self.run_into_wall(timecode)
            self.enable_event(EVENT_RUN_INTO_WALL)
        if event & EVENT_FOUND_COIN != 0 :
            self.disable_event(EVENT_FOUND_COIN)
            self.found_coin(timecode)
            self.enable_event(EVENT_FOUND_COIN)
        if event & EVENT_DETECTED_BY_FINDER != 0 :
            self.disable_event(EVENT_DETECTED_BY_FINDER)
            self.detected_by_finder(timecode)
            self.enable_event(EVENT_DETECTED_BY_FINDER)
        if event & EVENT_DETECTED_BY_RADAR != 0 :
            self.disable_event(EVENT_DETECTED_BY_RADAR)
            self.detected_by_radar(timecode)
            self.enable_event(EVENT_DETECTED_BY_RADAR)
        if event & EVENT_GOT_MESSAGE != 0 :
            self.disable_event(EVENT_GOT_MESSAGE)
            self.got_message(timecode)
            self.enable_event(EVENT_GOT_MESSAGE)
        self.last_message = msg_stack
        self.team = curr_team

    def convert_str2hex(self,str):
        hex=''
        for c in str:
            hex = hex + '%02x'%ord(c)
        return hex

    def char2int(self,c):
        r=0
        if ord(c) >= ord('0') and ord(c) <= ord('9'):
            r = ord(c) - ord('0')
        elif ord(c)>=ord('a') and ord(c)<=ord('f'):
            r = ord(c) - ord('a') + 10 
        elif ord(c)>=ord('A') and ord(c)<=ord('F'):
            r = ord(c) - ord('A') + 10 
        return r ;

    def convert_hex2str(self,hex):
        str=''
        d0=0
        d1=0
        cnt=0
        for c in hex:
            if c == ' ':
                continue 
            if cnt%2==0:
                d1 = self.char2int(c)
            else:
                d0 = self.char2int(c)
                str = str + chr(d1*16+d0)
            cnt=cnt+1
        return str

    def hit_by_bullet(self,timecode):
        return

    def run_into_turtle(self,timecode):
        return

    def run_into_donut(self,timecode):
        return

    def run_into_stone(self,timecode):
        return

    def run_into_wall(self,timecode):
        return

    def found_coin(self,timecode):
        return

    def detected_by_finder(self,timecode):
        return

    def detected_by_radar(self,timecode):
        return

    def got_message(self,timecode):
        return

    def home(self):
        self.ttl_send_command("C H")

    def rst(self):
        self.ttl_send_command("C") 

    def cf(self):
        self.ttl_send_command("C F")

    def clr(self):
        self.ttl_send_command("C A") 

    def fd(self,step):
        self.ttl_send_command("F %f"%step)

    def bk(self,step):
        self.ttl_send_command("B %f"%step) 

    def jump(self,x,y):
        self.ttl_send_command("J %f %f"%(x,y)) 

    def rjump(self,dx,dy):
        self.ttl_send_command("G %f %f"%(dx,dy)) 

    def rt(self,angle):
        self.ttl_send_command("T R %f"%angle) 

    def lt(self,angle):
        self.ttl_send_command("T L %f"%angle) 

    def east(self):
        self.ttl_send_command("T E") 

    def north(self):
        self.ttl_send_command("T N") 

    def west(self):
        self.ttl_send_command("T W") 

    def south(self):
        self.ttl_send_command("T S") 

    def lw(self,width):
        self.ttl_send_command("W %f"%width)

    def pd(self):
        self.ttl_send_command("P D") 

    def pu(self):
        self.ttl_send_command("P U") 

    def fill(self):
        self.ttl_send_command("L +") 

    def brush(self):
        self.ttl_send_command("L B") 

    def line(self):
        self.ttl_send_command("L -") 

    def point(self):
        self.ttl_send_command("L P") 

    def idle(self,t):
        sleep(t)

    def rf(self,range):
        self.ttl_send_command("E %f"%range)

    def say(self,something):
        str = self.convert_str2hex(something)
        self.ttl_send_command("S "+str) 

    def bcas(self,msg):
        str = self.convert_str2hex(msg)
        self.ttl_send_command("U "+str) 

    def tprintf(self, format, *args):
        msg = format%args
        self.say(msg)

#   NOT YET DONE
#   def tscanf(self,format):

    def fire(self):
        self.ttl_send_command("X") 

    def robot(self):
        self.ttl_send_command("R") 

    def donut(self):
        self.ttl_send_command("D") 

    doughnut=donut

    def coin(self):
        self.ttl_send_command("O X")

    def borrowcoin(self):
        self.ttl_send_command("O B")

    borrow_coin=borrowcoin
        
    def dropcoin(self):
        self.ttl_send_command("O S") 
        
    drop_coin=dropcoin

    def pickcoin(self):
        self.ttl_send_command("O R") 

    pick_coin=pickcoin

    def col(self, r, g, b):
        self.ttl_send_command("K %f %f %f"%(r,g,b)) 

    def bgc(self,r, g, b):
        self.ttl_send_command("H %f %f %f"%(r,g,b))

    def card(self,c,x,y):
        self.ttl_send_command("Z %d %f %f"%(c,x,y))

    def clrcard(self):
        self.ttl_send_command("Z 0")

    def bmode(self):
        self.ttl_send_command("! +")

    def gmode(self):
        self.ttl_send_command("! -")

    def nm(self,nickname):
        str = self.convert_str2hex(nickname)
        self.ttl_send_command("N "+str)

    def tm(self,team):
        self.ttl_send_command("I %d"%team)

    def q_id(self):
        self.ttl_send_command("Q M")
        ack = self.last_message.split(" ")
        return int(ack[0])

    def q_pos(self):
        self.ttl_send_command("Q P")
        ack = self.last_message.split(" ")
        return [float(ack[0]), float(ack[1])]

    def q_dir(self):
        self.ttl_send_command("Q D")
        ack = self.last_message.split(" ")
        return float(ack[0])

    q_heading=q_dir

    def q_alive(self):
        try:
            self.socket.send("Q A\n")
            self.last_message = self.socket.recv(BUFFSIZE)
        except IOError:
            return False
        if not self.last_message:
            return False
        else:
            return True

    def q_score(self):
        self.ttl_send_command("Q S")
        ack = self.last_message.split(" ")
        return int(ack[0])

    def q_nt(self):
        self.ttl_send_command("Q N")
        ack = self.last_message.split(" ")
        return int(ack[0])

    def q_coin(self):
        self.ttl_send_command("Q C")
        ack = self.last_message.split(" ")
        return int(ack[0])

    def q_mycoin(self):
        self.ttl_send_command("Q O")
        ack = self.last_message.split(" ")
        return int(ack[0])

    def q_radar(self):
        self.ttl_send_command("Q R")
        ack = self.last_message.split(" ")
        self.team = int(ack[1]) ;
        return float(ack[0])

    def q_radar_n(self,nn):
        self.ttl_send_command("Q R %d"%nn)
        ack = self.last_message.split(" ")
        self.team = int(ack[1]) ;
        return float(ack[0])

    def q_finder(self):
        self.ttl_send_command("Q F")
        ack = self.last_message.split(" ")
        self.team = int(ack[1]) ;
        return int(ack[0])

    def q_tm(self):
        return self.team

    def q_sonar(self):
        self.ttl_send_command("Q W")
        ack = self.last_message.split(" ")
        if int(ack[0]) > 0:
            left=True
        else:
            left=False
        if int(ack[1]) > 0:
            front=True
        else:
            front=False
        if int(ack[2]) > 0:
            right=True
        else:
            right=False
        return [left, front, right]

    def q_bcas(self):
        self.ttl_send_command("Q B")
        ack = self.last_message.split(" ")
        return [int(ack[0]), self.convert_hex2str(ack[1])]

    def q_time(self):
        self.ttl_send_command("Q T")
        ack = self.last_message.split(" ")
        return int(ack[0])

    def q_stat_card(self,c):
        self.ttl_send_command("Q Z S %d"%c)
        ack = self.last_message.split(" ")
        return int(ack[0])

    def q_pos_card(self,c):
        self.ttl_send_command("Q Z P %d"%c)
        ack = self.last_message.split(" ")
        return [float(ack[0]), float(ack[1])]
        
